use std::{collections::HashMap, fs::File, io::{BufReader, Read, Seek, SeekFrom}, iter::zip};

use bincode::{ErrorKind, Result};
use bitfield::bitfield;
use log::{debug, info};
use serde::Deserialize;

use super::{deserialize, deserialize_vec, index_to_location::IndexToLocation, table_directory::{FontTable, TableDirectoryRecord}};


#[derive(Debug, Deserialize)]
pub struct GlyphHeader {
	num_countours: i16,

	xmin: i16,
	ymin: i16,
	xmax: i16,
	ymax: i16
}

bitfield! {
	#[derive(Copy, Clone, Deserialize)]
	struct GlyphFlagData(u8);
	impl Debug;
	on_curve, _: 0;
	x_short, _: 1;
	y_short, _: 2;
	repeat, _: 3;
	x_same_or_positive, _: 4;
	y_same_or_positive, _: 5;
	overlap_simple, _: 6;
	reserved, _: 7;
}

#[derive(Debug)]
struct GlyphFlag {
	flags: GlyphFlagData,
	repeat: u8
}

#[derive(Debug)]
struct GlyphFlags<'a> {
	flags: &'a Vec<GlyphFlag>,
	current_item: usize,
	repetitions: u16
}

impl<'a> GlyphFlags<'a> {
	fn new(flags: &'a Vec<GlyphFlag>) -> GlyphFlags<'a> {
		GlyphFlags { flags, current_item: 0, repetitions: 0 }
	}
}

impl<'a> Iterator for GlyphFlags<'a> {
	type Item = GlyphFlagData;

	fn next(&mut self) -> Option<Self::Item> {
		if self.current_item >= self.flags.len() {
			return None;
		}

		let current = &self.flags[self.current_item];

		if self.repetitions < current.repeat as u16 {
			self.repetitions += 1;
			return Some(current.flags);
		}

		self.current_item += 1;
		if self.current_item >= self.flags.len() {
			return None;
		}

		let current = &self.flags[self.current_item];
		self.repetitions = 0;

		Some(current.flags)
		
	}
}

#[derive(Debug)]
pub struct GlyphPoint {
	x: i32,
	y: i32,

	on_curve: bool
}

#[derive(Debug)]
pub struct SimpleGlyph {
	header: GlyphHeader,
	instructions: Vec<u8>,

	points: Vec<GlyphPoint>
}

impl SimpleGlyph {
	fn read_flags<R: Read + Seek>(reader: &mut R, num_points: usize) -> Result<Vec<GlyphFlag>> {
		debug!("reading flags from position {}", reader.stream_position()?);
		let mut flags: Vec<GlyphFlag> = vec![];

		let mut points_processed: usize = 0;
		while points_processed < num_points {
			let flag: GlyphFlagData = deserialize(reader.by_ref())?;

			let repetitions: u8 = if flag.repeat() { deserialize(reader.by_ref())? } else { 0 };
			flags.push(GlyphFlag {
				flags: flag,
				repeat: repetitions
			});

			points_processed += repetitions as usize + 1;
			debug!("Flag: {flag:#?}, repeated {repetitions} times, logical elements processed: {points_processed}")
		}

		// debug!("glyph flags: {flags:#?}");
		Ok(flags)
	}

	fn read_coordinates<R: Read + Seek>(reader: &mut R, flags: GlyphFlags, is_y: bool) -> Result<Vec<(i32, bool)>> {
		let mut coordinates: Vec<(i32, bool)> = vec![];
		let mut current_coordinate = 0i32;

		for flag in flags {
			let short_vec = if is_y { flag.y_short() } else { flag.x_short() };
			let same_or_positive = if is_y { flag.y_same_or_positive() } else { flag.x_same_or_positive() };

			if short_vec {
				let sign: i32 = if same_or_positive { 1 } else { -1 };
				let delta = sign * (deserialize::<_, u8>(reader.by_ref())? as i32);

				current_coordinate += delta;
				coordinates.push((current_coordinate, flag.on_curve()));
			} else {
				if same_or_positive {
					coordinates.push((current_coordinate, flag.on_curve()));
				} else {
					let delta: i16 = deserialize(reader.by_ref())?;
					current_coordinate += delta as i32;
					coordinates.push((current_coordinate, flag.on_curve()));
				}
			}
		}

		Ok(coordinates)
	}

	fn new<R: Read + Seek>(header: GlyphHeader, reader: &mut R) -> Result<SimpleGlyph> {
		let end_points: Vec<u16> = deserialize_vec(reader, header.num_countours as usize)?;
		debug!("glyph end points: {end_points:#?}");

		let num_instructions: u16 = deserialize(reader.by_ref())?;
		let instructions: Vec<u8> = deserialize_vec(reader, num_instructions as usize)?;
		debug!("glyph instructions: {num_instructions}");

		let flags = SimpleGlyph::read_flags(reader, *end_points.last().unwrap() as usize)?;

		let x_coords = SimpleGlyph::read_coordinates(reader, GlyphFlags::new(&flags), false)?;
		let y_coords = SimpleGlyph::read_coordinates(reader, GlyphFlags::new(&flags), true)?;

		let coordinates = zip(x_coords.iter().cloned(), y_coords.iter().cloned())
			.map(|(x, y)| {
				GlyphPoint {
					x: x.0, y: y.0, on_curve: x.1
				}
			});

		// coordinates.for_each(|point| debug!("({}, {}, {})", point.x, point.y, if point.on_curve { "on" } else { "off" }));

		Ok(SimpleGlyph { header, instructions, points: coordinates.collect() })
	}
}

#[derive(Debug)]
pub enum GlyphData {
	Simple(SimpleGlyph),
	Composite
}

#[derive(Debug)]
pub struct GlyphTable {
	data_source: BufReader<File>,
	offset: u32,

	cache: HashMap<u16, GlyphData>
}

impl GlyphTable {
	pub fn get_glyph(&mut self, glyph_id: u16, location: &mut IndexToLocation) -> Result<GlyphData>{
		let Some(offset) = location.get_offset(glyph_id) else {
			return Err(ErrorKind::Custom("Failed to look up glyph location".into()).into());
		};

		debug!("glyph has offset {offset}");
		self.data_source.seek(SeekFrom::Start(self.offset as u64 + offset as u64))?;

		debug!("reading glyph data from position {}", self.offset as u64 + offset as u64);

		let header: GlyphHeader = deserialize(self.data_source.get_ref())?;
		debug!("glyph header: {header:#?}");

		if header.num_countours >= 0 {
			Ok(GlyphData::Simple(SimpleGlyph::new(header, self.data_source.by_ref())?))
		} else {
			todo!();
		}
	}
}

impl FontTable for GlyphTable {
	const TAG: &'static str = "glyf";
	type InitData = ();

	fn decode(reader: BufReader<File>, table: &TableDirectoryRecord, _: Option<Self::InitData>) -> bincode::Result<Self>
		where Self: Sized 
	{
		Ok(GlyphTable {
			data_source: reader,
			offset: table.offset,

			cache: HashMap::new()
		})
	}
}
